function [ Ys, W, I ] = MMI(Xs,k,Jp,d,W0,eta,nIter,useIC)
%
% Maximize mutual information between projected space and the class label
%
% Inputs:
%   Xs: original feature patterns of size [ D, N ]
%   k: pattern labels of size [ N, 1 ]
%   Jp: 
%   d: desired reduced dimension (must satisfy 1 <= d <= D)
%   W0: initial guess at the projection matrix of size [ D, d ]
% Outputs:
%   W: final projection matrix of size [ D, d ]
% 
% Written by:
% -- 
% John L. Weatherwax                2005-10-30
% 
% email: wax@alum.mit.edu
% 
% Please send comments and especially bug reports to the
% above email address.
% 
%-----

%MAX_ITER_NUM=20; % <- testing ...
MAX_ITER_NUM=nIter;   

%ETA = 1; % The learning rate
ETA = eta; 

%USE_IC = 1; % Which form of mutual information to use {0,1}
USE_IC = useIC;

%
% Code begins:

[ N, D ] = size(Xs); 
if( d>D ) error( 'Input value of d is too large' ); end

% Initialize theta (no rotation from initial projection matrix):
thetas = zeros( d*(D-d), 1 );
%thetas = mod( 2*pi*randn( d*(D-d), 1 ), 2*pi ); 

W = W0; 
Ys = Xs * W;
sgma = halfTwoFarthestPoints( Ys ); %sgma=10.0; 
no=0; 
fprintf('>s: no=%3d: sgma=%g\n',no,sgma);
I_old = -inf; 
for ni=1:MAX_ITER_NUM,
% $$$ ni=1; deltaTheta = inf;
% $$$ while( ni < 100 & norm(deltaTheta) > 0.001 )
% $$$   ni=ni+1; 
  %disp( [ 'Iteration #', num2str(ni), '...' ] ); 

  %disp( [ '   sgma = ',num2str(sgma) ] ); 
  
  [ dICdY, dITdY, IC, IT ] = computeInformationGradients(Ys,k,sgma,Jp);
  %fprintf( '   mmax(dICdY) = %g; mmax(dITdY) = %g\n',mmax(dICdY), mmax(dITdY) ); 
  if( ni>1 ) I_old = I; end
  if( USE_IC ) 
    dIdY = dICdY; I = IC; 
  else 
    dIdY = dITdY; I = IT; 
  end; 
  fprintf( 'I = %g\n', I ); 
  
  %dWdT = computeDWDTheta( thetas, d, D, W0 );  
  thetas = zeros( d*(D-d), 1 );
  dWdT = computeDWDTheta( thetas, d, D, W );
  dMIdT = dMIdTheta( dIdY, Xs, dWdT ); 
    
  % Plot this projection: 
  if( 1 ) 
    cs = { 'yo', 'g*', 'bd', 'rs', 'bx', 'm+' }; 
    figure; 
    for i=1:length(unique(k))
      inds=find(k==i);
      plot( -Ys(inds,1), +Ys(inds,2), cs{i} ); 
      %plot( +Ys(inds,1), -Ys(inds,2), cs{i} ); 
      %plot( -Ys(inds,1), -Ys(inds,2), cs{i} ); 
      if( i==1 ) hold on; end
    end
    axis( [ -3.5 3.5 -3.5 3.5 ] ); grid on; 
    title( ['Projected Data; I=',num2str(I)] ); drawnow; 
  end
  
  pause; 
  %keyboard; 
  
  % Find an "optimal" eta:
  etaMeth = 1; 
  switch etaMeth
   case 1
    % Perform very naive line search to find an "optimal" eta:
    % --first bracket our maximum by testing a great number of hypothesis:
    etaRng = 10.^[ -1:10 ]; 
    for ei=1:length(etaRng)
      thetasp = thetas + etaRng(ei) * dMIdT; 
      Wp = computeW( thetasp, d, D, W ); 
      Q = gram(Wp); Wp = Q; %clear Q; 
      Ys = Xs * Wp;
      [ VCY2, VC2Y2, VCY ] = computeInformationPotentials(Ys,k,Jp,sgma); 
      IT = VCY2 + VC2Y2 - 2*VCY;
      fprintf( '  %5d: ETA=%10g; IT=%20g', ei, etaRng(ei), IT ); 
      if( ei~=length(etaRng) ) fprintf('\n'); end
      Is(ei) = IT; 
    end
    % select the best eta: 
    [dum,ii]=max(Is); fprintf('          i_opt=%d\n',ii); 
    etaStr = etaRng(ii); 
   case 2
    % Levenberg-Marquardt Optimization: 
    eta = 0.95; 
    error( 'Not implemented...' ); 
   otherwise
    error( 'Unknown value of etaMeth' ); 
  end
  thetas = thetas + etaStr * dMIdT;   
  %W = computeW( thetas, d, D, W0 ); 
  W = computeW( thetas, d, D, W ); 
  Q = gram(W); W = Q; clear Q; 
  Ys = Xs * W;
  
  %fprintf( 'frac delta I = %g\n', abs(I_old-I)/I_old ); 
  if( abs(I_old-I)/I_old < 10^(-6) ) 
    sgma=sgma/2; 
  end
  
  if( 0 ) 
    cs = { 'ko', 'k*', 'kd', 'ks', 'kx', 'k+' }; 
    for i=1:length(unique(k))
      inds=find(k==i);
      plot( -Ys(inds,1), Ys(inds,2), cs{i} ); 
    end
    axis( [ -3.5 3.5 -3.5 3.5 ] ); grid on; 
    title( 'Projected Data' );    
  end
  if( 0 ) 
    % Plot each iteration: 
    cs = { 'yo', 'g*', 'bd', 'rs', 'bx', 'm+' }; 
    figure; 
    for i=1:length(unique(k))
      inds=find(k==i);
      plot( -Ys(inds,1), Ys(inds,2), cs{i} ); 
      if( i==1 ) hold on; end
    end
    axis( [ -3.5 3.5 -3.5 3.5 ] ); grid on; drawnow;
  end
  
  %keyboard;
end
